import java.util.*;
import java.io.*;
    
/**
    Program to read in strings from a file
    and insert them into a dictionary.
    
    @author Donald Chinn
    @version October 29, 2002
	@modified 11/9/04 by edhong to use standard Java collections API.
	                  also changed the set of implementations used.
    Flags:
      -b  binary search tree
      -v  AVL tree
      -r  Red/Black tree (Java TreeMap)
      -h  Hash table (Java HashMap, uses chaining)
	  -q  Hash table, (quadratic probing)
      -x  do nothing
*/
public class DictionaryApp {
    
    // static variables used to identify the
    // data structure/algorithm to use
    // (based on the command line argument)
    private static final int NOALG = 0;
    private static final int useBST = 1;
    private static final int useAVL = 2;
    private static final int useRedBlack = 3;
    private static final int useJavaHashTable = 4;
	private static final int useHashTableQuad = 5;
    private static final int useNothing = 6;
        
    /**
     * Return the next word (Unicode, big endian) in a file.
     * @param file  the file to read from
     * @return  the next word in the file, or null if there is no next word
     * A word is defined here as a consecutive sequence of
     * alphanumeric characters.
     */
    private static String getWord(DataInputStream file) {
        String currentString = null;
        char currentChar;
        try {
            // get rid of leading non-alphanumeric characters
            currentChar = file.readChar();
            while (!Character.isLetterOrDigit(currentChar)) {
                currentChar = file.readChar();
            }
        }
        catch (EOFException eof) {
            return null;
        }
        catch (IOException io) {
            return null;
        }
        
        currentString = new String(currentChar + "");
    
        try {
            // read in the next word
            currentChar = file.readChar();
            while (Character.isLetterOrDigit(currentChar)) {
                currentString += currentChar;
                currentChar = file.readChar();
            }
        }
        catch (EOFException eof) {
            return currentString;
        }
        catch (IOException io) {
            return currentString;
        }
        return currentString;
    }

    /**
     * The driver method for the word counting application.
     */
    public static void main(String[] args) {
        boolean error = false;

        // timer variables
        long totalTime = 0;
        long startTime = 0;
        long finishTime = 0;
        int numInsertions = 0;

        int whichAlgorithm = NOALG;
    
        DataInputStream infile = null;
            
        // Handle command line arguments.
        // Usage:  args[0] -[bvshx] input_filename
        // Options:
        //   -b  use a standard binary search tree
        //   -v  use an AVL tree
        //   -r  use a red-black tree from Java's TreeMap
		//   -h  us a Hash table with chaining from Java's HashMap.
	    //   -q  use a hash table with quadratic probing
        //   -x  no data structure (just read in the file)
        
        if ((args.length < 2) || (args.length > 2)) {
            System.out.println("Argument usage: -[bvrhqx] infile");
            error = true;
        } else {
            // figure out which option was chosen
            if (args[0].charAt(0) == '-') {
                switch (args[0].charAt(1)) {
                    case 'b':
                        whichAlgorithm = useBST;
                        break;
    
                    case 'v':
                        whichAlgorithm = useAVL;
                        break;
    
                    case 'r':
                        whichAlgorithm = useRedBlack;
                        break;
            
                    case 'h':
                        whichAlgorithm = useJavaHashTable;
                        break;

                    case 'q':
                        whichAlgorithm = useHashTableQuad;
                        break;
                    
                    case 'x':
                        whichAlgorithm = useNothing;
                        break;
            
                    default:
                        System.out.print ("Usage: ");
                        System.out.println("-" + args[0].charAt(1) + " is not a valid option.");
                        error = true;
                        break;
                }
    
                // Get the input filename
                try {
                    infile = new DataInputStream(new FileInputStream (args[1]));
                }
                catch (IOException ioexception) {
                  System.out.println("Error: Could not open " + args[1] + ".");
                  error = true;
                }
            } else {
                System.out.println("Argument usage: -[bvshx] filename");
                error = true;
            }
        }
    
        if (!error) {
			String currentWord;
			Map datamap = null;
			SimpleMap datasimplemap = null;
			// use the correct data structure. 
			switch (whichAlgorithm) {
                    case useBST:
					    datasimplemap = (SimpleMap) new BinarySearchTreeMap();
						break;
        
                    case useAVL:
						datasimplemap = (SimpleMap) new AvlTreeMap();
                        break;
						
					case useRedBlack:
						datamap = (Map) new TreeMap();
                        break;
            
                    case useJavaHashTable:
						datamap = (Map) new HashMap();
                        break;

                    case useHashTableQuad:
					    System.out.println("Error, Quadratic Probing Hashtable not yet implemented");
//						datasimplemap = (SimpleMap) new QuadraticProbingHashTableMap();
						break;
			}

			Integer counter = null;
            // start the timer
            Date startDate = new Date();
            startTime = startDate.getTime();
          
            while ((currentWord = DictionaryApp.getWord(infile)) != null) {
                switch (whichAlgorithm) {
                    case useBST:
                    case useAVL:
                    case useHashTableQuad:
						counter = (Integer) datasimplemap.get(currentWord);
						if (counter == null) {
						    // first time currentWord has been seen; set counter to 1.
							counter = new Integer(1);
						}
						else {
						    // currentword has been seen before; increment count
							counter = new Integer(counter.intValue()+1);
						}
					    datasimplemap.put(currentWord, counter);
						numInsertions++;
						break;
						
					case useRedBlack:
                    case useJavaHashTable:
						counter = (Integer) datamap.get(currentWord);
						if (counter == null) {
						    // first time currentWord has been seen; set counter to 1.
							counter = new Integer(1);
						}
						else {
						    // currentword has been seen before; increment count
							counter = new Integer(counter.intValue()+1);
						}
					    datamap.put(currentWord, counter);
						numInsertions++;
						break;
						
					case useNothing:
                        // do nothing
                        numInsertions++;
                        break;
                }
            }

            // stop the timer
            Date finishDate = new Date();
            finishTime = finishDate.getTime();
            totalTime += (finishTime - startTime);

            // if the data structure was a binary search tree, then
            // print out the first 50 words (alphabetically close to A)
			// and the most frequently occuring words (top 50)
            if (whichAlgorithm == useBST || whichAlgorithm == useAVL) {
				BinarySearchTreeMap bstmap = (BinarySearchTreeMap) datasimplemap;
				System.out.println("First 10 words");
				bstmap.printFirst10();
				System.out.println("Most Frequent 10 words");
				bstmap.print10MostFrequent();
            }
            
            System.out.println("** Results for " + args[0]  + " option on file ");
            System.out.println("    " + args[1]);
            System.out.println("Time to do insertions: ");
            System.out.println(totalTime + " milliseconds.");
            System.out.println("Number of insertions: " + numInsertions);
    
        }
    
        // Close the file stream.
        if (!error) {
            try {
                infile.close();
            }
            catch (IOException ioexception) {
                System.out.println("Error: Could not close " + args[1] + ".");
                error = true;
            }
        }
    
        System.out.println("DictionaryApp is done. **");
    }
    
}
